@@ -31,6 +31,7 @@
 #include <signal.h> /* kill */
 #include <sys/stat.h> /* struct stat */
 #include <sys/wait.h>
+#include <sys/sem.h> /* SysV semaphore API */
 
 #include <dirent.h>
 
@@ -51,7 +52,7 @@
 #endif
 
 #ifndef SIGCLD
-    #define SIGCLD SIGCHLD /* not defined on OS X */
+    #define SIGCLD SIGCHLD /* not defined on macOS */
 #endif
 
 #ifndef sighandler_t
@@ -491,15 +492,38 @@ _ignoring_signal_handler(int signum)
 static void
 _signal_handler(int signum)
 {
+    int original_errno = errno; /* Store original errno */
+    pid_t child_pid = global_pyi_ctx->child_pid; /* Read from volatile global variable */
+
     /* Forward signal to the child. Avoid generating debug messages, as
      * functions involved are generally not signal safe. Furthermore, it
      * may result in endless spamming of SIGPIPE, as reported and
      * diagnosed in #5270. */
-    if (!global_pyi_ctx->child_pid) {
-        /* No-op if child process does not exist (yet or anymore) */
+
+#if defined(LAUNCH_DEBUG)
+    global_pyi_ctx->signal_forward_all++;
+#endif
+
+    if (child_pid <= 0) {
+        /* No-op if child process does not exist (yet or anymore), or
+         * if fork() returned -1 due to error. */
+#if defined(LAUNCH_DEBUG)
+        global_pyi_ctx->signal_forward_noop++;
+#endif
         return;
     }
-    kill(global_pyi_ctx->child_pid, signum);
+
+#if defined(LAUNCH_DEBUG)
+    if (kill(child_pid, signum) == 0) {
+        global_pyi_ctx->signal_forward_ok++;
+    } else {
+        global_pyi_ctx->signal_forward_error++;
+    }
+#else
+    kill(child_pid, signum);
+#endif
+
+    errno = original_errno; /* Restore original errno */
 }
 
 /* Start frozen application in a subprocess. The frozen application runs
@@ -507,7 +531,7 @@ _signal_handler(int signum)
 int
 pyi_utils_create_child(struct PYI_CONTEXT *pyi_ctx)
 {
-    pid_t pid = 0;
+    pid_t child_pid = 0;
     int rc = 0;
     int wait_rc = -1;
 
@@ -518,40 +542,49 @@ pyi_utils_create_child(struct PYI_CONTEXT *pyi_ctx)
     sighandler_t signal_handler;
     int signum;
 
-    /* Install signal handlers to either forward received signals to the
-     * child process, or ignore them (effectively blocking them).
+    /* Create semaphore for synchronizing child and parent; we need to
+     * ensure that the child starts executing user's python code only
+     * *after* the parent has fully completed the `fork()` call *and* stored
+     * the child process ID to `pyi_ctx->child_pid` for the forwarding
+     * signal handlers (as well as installed the said handlers).
+     * Failing to do so leads to sporadic failures in our signal handling
+     * test (`test_onefile_signal_handling`) when performed under heavy
+     * CPU load.
      *
-     * NOTE: we install signal handlers *before* forking the child process
-     * in order to prevent sporadic failures of our signal handling test
-     * (`test_onefile_signal_handling`); there, we signal the parent
-     * process from the child, and depending on the scenario (forward/block)
-     * expect child to either receive the forwarded signal or not receive
-     * the signal. If the signal handlers are installed after the `fork()`
-     * and `execvp()`, it might happen (in CPU contention scenarios) that
-     * the child proceeds with execution while the parent stalls for a bit,
-     * and thus child ends up signalling the parent before the latter has
-     * set up its signal handlers.
-     *
-     * In practice, this early setup should not make any difference; while
-     * the signal handlers are inherited when the child process is forked,
-     * the subsequent program re-execution via execvp() ends up resetting
-     * them. */
-    if (pyi_ctx->ignore_signals) {
-        PYI_DEBUG("LOADER: registering signal handlers to ignore received signals.\n");
-        signal_handler = &_ignoring_signal_handler;
-    } else {
-        PYI_DEBUG("LOADER: registering signal handlers to forward received signals to child.\n");
-        signal_handler = &_signal_handler;
+     * The first choice for API would be un-named POSIX signals, but this
+     * is not supported on macOS. So we use the old SysV signals API... */
+
+    /* Argument for semctl(); according to API, we need to provide our
+     * own union definition... */
+    union {
+        int val;
+        struct semid_ds *buf;
+        unsigned short *array;
+    } sem_arg;
+
+    struct sembuf sem_op;  /* Argument for semop() */
+    int sem_id = -1;  /* Semaphore (array) ID */
+
+    /* Create semaphore */
+    sem_id = semget(IPC_PRIVATE, 1, 0660 | IPC_CREAT | IPC_EXCL);
+    if (sem_id < 0) {
+        /* Allow semaphore creation to fail, for whatever reason, and
+         * disable parent/child sync in that case. Not having the sync
+         * should be of little consequence outside of PyInstaller's own
+         * POSIX signal test suite, so it should not be considered fatal.
+         * For example, under cygwin, semget() is only available when
+         * cygserver is running, which might not always be the case.
+         * There might be other failure modes on other POSIX systems. */
+        PYI_DEBUG("LOADER: failed to create sync semaphore (errno %d) - disabling sync.\n", errno);
     }
 
-    for (signum = 0; signum < num_signals; ++signum) {
-        /* Don't mess with SIGCHLD/SIGCLD; it affects our ability
-         * to wait() for the child to exit. Similarly, do not change
-         * don't change SIGTSP handling to allow Ctrl-Z */
-        if (signum == SIGCHLD || signum == SIGCLD || signum == SIGTSTP) {
-            continue;
+    /* Initialize semaphore's value to 0 (= locked/acquired) */
+    if (sem_id >= 0) {
+        sem_arg.val = 0;
+        if (semctl(sem_id, 0, SETVAL, sem_arg) < 0) {
+            PYI_PERROR("semctl", "Failed to initialize sync semaphore!\n");
+            goto cleanup;
         }
-        signal(signum, signal_handler);
     }
 
     /* macOS: Apple Events handling */
@@ -577,20 +610,32 @@ pyi_utils_create_child(struct PYI_CONTEXT *pyi_ctx)
 #endif
 
     /* Fork the child process. */
-    pid = fork();
-    if (pid < 0) {
+    child_pid = fork();
+    if (child_pid < 0) {
         PYI_WARNING("LOADER: failed to fork child process: %s\n", strerror(errno));
         goto cleanup;
     }
 
     /* Child code. */
-    if (pid == 0) {
+    if (child_pid == 0) {
         /* Replace process by starting a new application. */
         /* If modified arguments (pyi_ctx->pyi_argv) are available, use
          * those. Otherwise, use the original pyi_ctx->argv. */
         char *const *argv = (pyi_ctx->pyi_argv != NULL) ? pyi_ctx->pyi_argv : pyi_ctx->argv;
         const int argc = (pyi_ctx->pyi_argv != NULL) ? pyi_ctx->pyi_argc : pyi_ctx->argc;
 
+        /* Wait on the sync semaphore */
+        if (sem_id >= 0) {
+            PYI_DEBUG("LOADER: waiting on sync semaphore...\n");
+            sem_op.sem_num = 0;
+            sem_op.sem_op = -1; /* P/decrement (= acquire semaphore) */
+            sem_op.sem_flg = 0;
+            if (semop(sem_id, &sem_op, 1) < 0) {
+                PYI_PERROR("semop", "Failed to wait on sync semaphore!\n");
+            }
+        }
+
+        /* Modify the LISTEN_PID environment variable, if necessary */
         if (_pyi_set_systemd_env() != 0) {
             PYI_WARNING("LOADER: application is started by systemd socket, but we cannot set proper LISTEN_PID on it.\n");
         }
@@ -626,19 +671,50 @@ pyi_utils_create_child(struct PYI_CONTEXT *pyi_ctx)
         /* NOTREACHED */
     }
 
-    /* From here to end-of-function is parent code (since the child exec'd).
-     * The exception is the `cleanup` block that frees argv_pyi; in the child,
-     * wait_rc is -1, so the child exit code checking is skipped. */
+    /* From here to end-of-function is parent code (since the child exec'd,
+     * or exited on exec failure). */
+    PYI_DEBUG("LOADER: forked child process with PID: %d\n", child_pid);
 
-    /* Store the child PID. */
-    pyi_ctx->child_pid = pid;
+    /* Store child PID to context structure for use in forwarding signal
+     * handler (as well as Apple event handler on macOS). */
+    pyi_ctx->child_pid = child_pid;
+
+    /* Install signal handlers to either forward received signals to the
+     * child process, or ignore them (effectively blocking them). */
+    if (pyi_ctx->ignore_signals) {
+        PYI_DEBUG("LOADER: registering signal handlers to ignore received signals.\n");
+        signal_handler = &_ignoring_signal_handler;
+    } else {
+        PYI_DEBUG("LOADER: registering signal handlers to forward received signals to child.\n");
+        signal_handler = &_signal_handler;
+    }
+
+    for (signum = 0; signum < num_signals; ++signum) {
+        /* Don't mess with SIGCHLD/SIGCLD; it affects our ability
+         * to wait() for the child to exit. Similarly, do not change
+         * don't change SIGTSP handling to allow Ctrl-Z */
+        if (signum == SIGCHLD || signum == SIGCLD || signum == SIGTSTP) {
+            continue;
+        }
+        signal(signum, signal_handler);
+    }
+
+    if (sem_id >= 0) {
+        PYI_DEBUG("LOADER: signalling the sync semaphore...\n");
+        sem_op.sem_num = 0;
+        sem_op.sem_op = 1; /* V/increment (= release semaphore) */
+        sem_op.sem_flg = 0;
+        if (semop(sem_id, &sem_op, 1) < 0) {
+            PYI_PERROR("semop", "Failed to signal the sync semaphore!\n");
+        }
+    }
 
 #if defined(__APPLE__) && defined(WINDOWED)
     /* macOS: forward events to child */
     do {
         /* The below loop will iterate about once every second on Apple,
          * waiting on the event queue most of that time. */
-        wait_rc = waitpid(pyi_ctx->child_pid, &rc, WNOHANG);
+        wait_rc = waitpid(child_pid, &rc, WNOHANG);
         if (wait_rc == 0) {
             /* Check if we have a pending event that we need to forward... */
             if (pyi_apple_has_pending_event(pyi_ctx->ae_ctx)) {
@@ -664,7 +740,7 @@ pyi_utils_create_child(struct PYI_CONTEXT *pyi_ctx)
     /* Uninstall event handlers */
     pyi_apple_uninstall_event_handlers(&pyi_ctx->ae_ctx);
 #else
-    wait_rc = waitpid(pyi_ctx->child_pid, &rc, 0);
+    wait_rc = waitpid(child_pid, &rc, 0);
 #endif
 
     /* Reset stored child PID - this aims to immediately turn forwarding
@@ -681,7 +757,27 @@ pyi_utils_create_child(struct PYI_CONTEXT *pyi_ctx)
         signal(signum, SIG_DFL);
     }
 
+    /* Display statistics from forwarding signal-handler. */
+#if defined(LAUNCH_DEBUG)
+    if (!pyi_ctx->ignore_signals) {
+        PYI_DEBUG(
+            "LOADER: signal forwarding statistics: all=%u, ok=%u, err=%u, noop=%u\n",
+            pyi_ctx->signal_forward_all,
+            pyi_ctx->signal_forward_ok,
+            pyi_ctx->signal_forward_error,
+            pyi_ctx->signal_forward_noop
+        );
+    }
+#endif
+
 cleanup:
+    /* Destroy the sync semaphore (if available) */
+    if (sem_id >= 0) {
+        if (semctl(sem_id, 0, IPC_RMID) < 0) {
+            PYI_WARNING("LOADER: failed to destroy sync semaphore (errno %d)!\n", errno);
+        }
+    }
+
     /* Clean up the modified copy of command-line arguments (currently
      * applicable only to macOS windowed bootloader builds). */
 #if defined(__APPLE__) && defined(WINDOWED)
