@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 #-----------------------------------------------------------------------------
 # Copyright (c) 2005-2023, PyInstaller Development Team.
 #
@@ -16,11 +15,13 @@ import subprocess
 
 import pytest
 
+from PyInstaller.utils.tests import onefile_only
+
 
 @pytest.mark.darwin
 @pytest.mark.linux
 @pytest.mark.parametrize('forward_signals', [True, False], ids=['forward', 'ignore'])
-@pytest.mark.parametrize('pyi_builder', ['onefile'], indirect=True)  # Run only in onefile mode.
+@onefile_only
 def test_onefile_signal_handling(pyi_builder, forward_signals):
     # Build the test program. The `pyi_builder.test_soruce` also runs the built program, but since no arguments are
     # passed via command-line, this program run is a no-op.
@@ -66,11 +67,12 @@ def test_onefile_signal_handling(pyi_builder, forward_signals):
         # Wait for signal to be delivered or the specified timeout interval to pass.
         start_time = time.time()
         while True:
-            if time.time() - start_time >= timeout:
+            elapsed_time = time.time() - start_time
+            if elapsed_time >= timeout:
                 print(f"Signal not received within {timeout} seconds!", file=sys.stderr)
                 break
             if return_code != 0:
-                print("Signal received!", file=sys.stderr)
+                print(f"Signal received! Elapsed time: {elapsed_time:.2f} seconds.", file=sys.stderr)
                 break
             time.sleep(0.1)  # 100 ms steps
 
